import * as fs from "fs/promises";
import { URL } from "url";
import * as path from "path";

/**
 * Resource class for handling file and URL paths.
 */
export class Resource {
  /**
   * Creates a new Resource instance.
   * @param path - The file or URL path.
   */
  constructor(public readonly path: string) {}

  /**
   * Returns the string representation of the Resource instance.
   * @returns The path of the Resource.
   */
  toString(): string {
    return this.path;
  }

  /**
   * Checks if the path is a valid URL.
   * @return True if the path is a valid URL, false otherwise.
   */
  get isUrl(): boolean {
    try {
      new URL(this.path);
      return true;
    } catch {
      return false;
    }
  }

  /**
   * Gets the name of the file or URL.
   * @return The name of the file or URL.
   */
  get name(): string {
    return this.isUrl
      ? new URL(this.path).pathname.split("/").pop() || ""
      : path.basename(this.path);
  }

  /**
   * Gets the parent directory of the file or URL.
   * @return The parent directory of the file or URL.
   */
  get parent(): Resource {
    if (this.isUrl) {
      const url = new URL(this.path);
      url.pathname = path.dirname(url.pathname);
      return new Resource(url.toString());
    }
    return new Resource(path.dirname(this.path));
  }

  /**
   * Gets the file extension of the file or URL.
   * @return The file extension of the file or URL.
   */
  get ext(): string {
    return this.isUrl
      ? path.extname(new URL(this.path).pathname)
      : path.extname(this.path);
  }

  /**
   * Checks if the file or URL exists.
   * @return True if the file or URL exists, false otherwise.
   */
  async exists(): Promise<boolean> {
    if (this.isUrl) {
      const response = await fetch(this.path);
      if (!response.ok) return false;
      return true;
    } else {
      return fs.exists(this.path);
    }
  }

  /**
   * Fetches the contents of the file or URL.
   * @return The contents of file the Resource points to.
   */
  async fetchContents(): Promise<string> {
    if (this.isUrl) {
      const response = await fetch(this.path);
      if (!response.ok)
        throw new Error(`Failed to fetch ${this.path}: ${response.statusText}`);
      return await response.text();
    }
    return fs.readFile(this.path, { encoding: "utf8" });
  }

  /**
   * Joins the Resource with other segments to create a new Resource.
   * @param segments - The segments to join with the Resource.
   * @return A new Resource instance with the joined path.
   */
  join(...segments: string[]) {
    if (this.isUrl) {
      const url = new URL(this.path);
      url.pathname = path.join(url.pathname, ...segments);
      return new Resource(url.toString());
    }
    return new Resource(path.join(this.path, ...segments));
  }
}
